package main

import (
	"contract-case/log"
	"contract-case/operator"
	"contract-case/tool"
	"contract-case/util"
	"github.com/ethereum/go-ethereum/core/types"
)

func main() {
	log.InitLog(log.LogConfig{Path: "logs", Level: "debug", Save: 3})
	txArr := make([]*types.Transaction, 0)
	verifyFuncArr := make([]func() bool, 0)

	// 部署合约
	if tool.Cfg.IsDeployContract {
		txArr = operator.DeployContract(tool.Cfg.RpcNode, txArr)
	}

	// 初始化From账户以及合约账户
	if tool.Cfg.IsInitAccountBalance {
		txArr = operator.InitCaseAccount(tool.Cfg.RpcNode, txArr)
	}

	// 转账交易case
	if len(tool.Cfg.TransferType) != 0 {
		accIndex := tool.Cfg.AccStartIndex
		nonceMap := util.GetAccountNonce(tool.Cfg.RpcNode)
		arrFormat := tool.ParseAccountConfig("./config/account.json")
		parseContractConfig := tool.ParseContractConfig("./config/contractConfig.json")
		for _, caseType := range tool.Cfg.TransferType {
			tranList, verifyFunc := operator.GetTranFunc(caseType, accIndex, tool.Cfg.EveryCaseTxCount, nonceMap, arrFormat, tool.Cfg.RpcNode, parseContractConfig)
			if tranList == nil || verifyFunc == nil {
				break
			}
			txArr = append(txArr, tranList...)
			verifyFuncArr = append(verifyFuncArr, verifyFunc)
			if tool.Cfg.IsUseNotSameAcc {
				accIndex += 9
			}
		}
	}

	// 发送整合后的交易
	if len(txArr) > 0 {
		operator.SendTransaction(txArr)
		util.GetTxReceipt(tool.Cfg.RpcNode, txArr)
		for _, vf := range verifyFuncArr {
			if !vf() {
				log.Error("Verify failed")
			}
		}
	}

	if tool.Cfg.IsGetAccBalance {
		util.GetAccountBalance(tool.Cfg.RpcNode)
	}
}
